from typing import *
import datetime
import royalnet.utils as ru
from royalnet.constellation.api import *
from sqlalchemy import and_
from ..tables.tokens import Token


class ApiTokenPasswdStar(ApiStar):
    path = "/api/token/passwd/v1"

    methods = ["PUT"]

    summary = "Change Royalnet password for an user."

    tags = ["token"]

    parameters = {
        "new_password": "The password you want to set."
    }

    requires_auth = True

    async def api(self, data: ApiData) -> ru.JSON:
        TokenT = self.alchemy.get(Token)
        token = await data.token()
        user = token.user
        user.set_password(data["new_password"])
        tokens: List[Token] = await ru.asyncify(
            data.session
                .query(self.alchemy.get(Token))
                .filter(
                    and_(
                        TokenT.user == user,
                        TokenT.expiration >= datetime.datetime.now()
                    ))
                .all
        )
        for t in tokens:
            if t.token != token.token:
                t.expired = True
        await data.session_commit()
        return {
            "revoked_tokens": len(tokens) - 1
        }



def escape(string: str) -> str:
    """Escape a string to be sent through Discord, and format it using RoyalCode.

    Warning:
        Currently escapes everything, even items in code blocks."""
    return string.replace("*", "\\*") \
        .replace("_", "\\_") \
        .replace("`", "\\`") \
        .replace("[b]", "**") \
        .replace("[/b]", "**") \
        .replace("[i]", "_") \
        .replace("[/i]", "_") \
        .replace("[u]", "__") \
        .replace("[/u]", "__") \
        .replace("[c]", "`") \
        .replace("[/c]", "`") \
        .replace("[p]", "```") \
        .replace("[/p]", "```")



import typing
import re


def andformat(coll: typing.Collection[str], middle=", ", final=" and ") -> str:
    """Convert a collection (such as a :class:`list`) to a :class:`str` by adding ``final`` between the last two
    elements and ``middle`` between the others.

    Args:
        coll: the input collection.
        middle: the :class:`str` to be added between the middle elements.
        final: the :class:`str` to be added between the last two elements.

    Returns:
        The resulting :py:class:`str`.

    Examples:
        ::

            >>> andformat(["Steffo", "Kappa", "Proto"])
            "Steffo, Kappa and Proto"

            >>> andformat(["Viktya", "Sensei", "Cate"], final=" e ")
            "Viktya, Sensei e Cate"

            >>> andformat(["Paltri", "Spaggia", "Gesù", "Mallllco"], middle="+", final="+")
            "Paltri+Spaggia+Gesù+Mallllco"
    """
    result = ""
    for index, item in enumerate(coll):
        result += item
        if index == len(coll) - 2:
            result += final
        elif index != len(coll) - 1:
            result += middle
    return result


def underscorize(string: str) -> str:
    """Replace all non-word characters in a :class:`str` with underscores.

    It is particularly useful when you want to use random strings from the Internet as filenames.
    
    Parameters:
        string: the input string.
    
    Returns:
        The resulting string.

    Example:
        ::

            >>> underscorize("LE EPIC PRANK [GONE WRONG!?!?]")
            "LE_EPIC_PRANK__GONE_WRONG_____"

    """
    return re.sub(r"\W", "_", string)


def ytdldateformat(string: typing.Optional[str], separator: str = "-") -> str:
    """Convert the date :class:`str` returned by `youtube_dl <https://ytdl-org.github.io/youtube-dl/index.html>`_  into
     the ``YYYY-MM-DD`` format.
    
    Parameters:
        string: the input :class:`str`, in the ``YYYYMMDD`` format used by youtube_dl.
        separator: the :class:`str` to add between the years, the months and the days. Defaults to ``"-"``.
        
    Returns:
        The resulting :class:`str` in the new format.

    Example:
        ::

            >>> ytdldateformat("20111111")
            "2011-11-11"

            >>> ytdldateformat("20200202", separator=".")
            "2020.02.02"

    """
    if string is None:
        return ""
    return f"{string[0:4]}{separator}{string[4:6]}{separator}{string[6:8]}"


def numberemojiformat(li: typing.Collection[str]) -> str:
    """Convert a collection to a string with one item on every line numbered with emojis.

    Parameters:
        li: the list to convert.

    Returns:
        The resulting Unicode string.

    Examples:
        Cannot be displayed, as Sphinx does not render emojis properly.
    """
    number_emojis = ["1️⃣", "2️⃣", "3️⃣", "4️⃣", "5️⃣", "6️⃣", "7️⃣", "8️⃣", "9️⃣", "🔟"]
    extra_emoji = "*️⃣"
    result = ""
    for index, element in enumerate(li):
        try:
            result += f"{number_emojis[index]} {element}\n"
        except IndexError:
            result += f"{extra_emoji} {element}\n"
    return result


def ordinalformat(number: int) -> str:
    """Convert a :class:`int` to the corresponding English ordinal :class:`str`.

    Parameters:
        number: the number to convert.

    Returns:
        The corresponding English `ordinal numeral <https://en.wikipedia.org/wiki/Ordinal_numeral>`_.

    Examples:
        ::

            >>> ordinalformat(1)
            "1st"
            >>> ordinalformat(2)
            "2nd"
            >>> ordinalformat(11)
            "11th"
            >>> ordinalformat(101)
            "101st"
            >>> ordinalformat(112)
            "112th"
            >>> ordinalformat(0)
            "0th"
    """
    if 10 <= number % 100 < 20:
        return f"{number}th"
    if number % 10 == 1:
        return f"{number}st"
    elif number % 10 == 2:
        return f"{number}nd"
    elif number % 10 == 3:
        return f"{number}rd"
    return f"{number}th"



# Imports go here!
from .version import VersionCommand

# Enter the commands of your Pack here!
available_commands = [
    VersionCommand,
]

# Don't change this, it should automatically generate __all__
__all__ = [command.__name__ for command in available_commands]



"""The subpackage providing all classes related to Royalnet commands."""

from .commandinterface import CommandInterface
from .command import Command
from .commanddata import CommandData
from .commandargs import CommandArgs
from .event import Event
from .errors import \
    CommandError, InvalidInputError, UnsupportedError, ConfigurationError, ExternalError, UserError, ProgramError
from .keyboardkey import KeyboardKey

__all__ = [
    "CommandInterface",
    "Command",
    "CommandData",
    "CommandArgs",
    "CommandError",
    "InvalidInputError",
    "UnsupportedError",
    "ConfigurationError",
    "ExternalError",
    "UserError",
    "ProgramError",
    "Event",
    "KeyboardKey",
]



import discord


class FileAudioSource(discord.AudioSource):
    """A :py:class:`discord.AudioSource` that uses a :py:class:`io.BufferedIOBase` as an input instead of memory.

    The stream should be in the usual PCM encoding.

    Warning:
        This AudioSource will consume (and close) the passed stream."""

    def __init__(self, file):
        """Create a FileAudioSource.

        Arguments:
            file: the file to be played back."""
        self.file = file
        self._stopped = False

    def __repr__(self):
        if self.file.seekable():
            return f"<{self.__class__.__name__} @{self.file.tell()}>"
        else:
            return f"<{self.__class__.__name__}>"

    def is_opus(self):
        """This audio file isn't Opus-encoded, but PCM-encoded.

        Returns:
            ``False``."""
        return False

    def stop(self):
        """Stop the FileAudioSource. Once stopped, a FileAudioSource will immediatly stop reading more bytes from the
        file."""
        self._stopped = True

    def read(self):
        """Reads 20ms of audio.

        If the stream has ended, then return an empty :py:class:`bytes`-like object."""
        data: bytes = self.file.read(discord.opus.Encoder.FRAME_SIZE)
        # If there is no more data to be streamed
        if self._stopped or len(data) != discord.opus.Encoder.FRAME_SIZE:
            # Return that the stream has ended
            return b""
        return data



class HeraldError(Exception):
    """A generic :mod:`royalnet.herald` error."""


class LinkError(HeraldError):
    """An error for something that happened in a :class:`Link`."""


class ServerError(HeraldError):
    """An error for something that happened in a :class:`Server`."""


class ConnectionClosedError(LinkError):
    """The :py:class:`Link`'s connection was closed unexpectedly. The link can't be used anymore."""


class InvalidServerResponseError(LinkError):
    """The :py:class:`Server` sent invalid data to the :class:`Link`."""



"""A :class:`Serf` implementation for Telegram.

It requires (obviously) the ``telegram`` extra to be installed.

Install it with: ::

    pip install royalnet[telegram]

"""

from .escape import escape
from .telegramserf import TelegramSerf

__all__ = [
    "escape",
    "TelegramSerf"
]



from typing import *
from starlette.requests import Request
from starlette.responses import Response
from royalnet.commands import CommandInterface

if TYPE_CHECKING:
    from .constellation import Constellation


class Star:
    """A Star is a class representing a part of the website.

    It shouldn't be used directly: please use :class:`PageStar` and :class:`ExceptionStar` instead!"""
    def __init__(self, interface: CommandInterface):
        self.interface: CommandInterface = interface

    async def page(self, request: Request) -> Response:
        """The function generating the :class:`~starlette.Response` to a web :class:`~starlette.Request`.

        If it raises an error, the corresponding :class:`ExceptionStar` will be used to handle the request instead."""
        raise NotImplementedError()

    @property
    def constellation(self) -> "Constellation":
        """A shortcut for the :class:`Constellation`."""
        return self.interface.constellation

    @property
    def alchemy(self):
        """A shortcut for the :class:`~royalnet.alchemy.Alchemy` of the :class:`Constellation`."""
        return self.interface.constellation.alchemy

    # noinspection PyPep8Naming
    @property
    def Session(self):
        """A shortcut for the :class:`~royalnet.alchemy.Alchemy` :class:`Session` of the :class:`Constellation`."""
        return self.interface.constellation.alchemy.Session

    @property
    def session_acm(self):
        """A shortcut for :func:`.alchemy.session_acm` of the :class:`Constellation`."""
        return self.interface.constellation.alchemy.session_acm

    @property
    def config(self) -> Dict[str, Any]:
        """A shortcut for the Pack configuration of the :class:`Constellation`."""
        return self.interface.config

    def __repr__(self):
        return f"<{self.__class__.__qualname__}>"



import asyncio as aio
import logging
import warnings
from typing import *
import royalnet.backpack as rb
import royalnet.commands as rc
from royalnet.utils import asyncify
from royalnet.serf import Serf
from .escape import escape
from .voiceplayer import VoicePlayer


try:
    import discord
except ImportError:
    discord = None

try:
    from sqlalchemy.orm.session import Session
except ImportError:
    Session = None

try:
    from royalnet.herald import Config as HeraldConfig
except ImportError:
    HeraldConfig = None

log = logging.getLogger(__name__)


class DiscordSerf(Serf):
    """A :class:`Serf` that connects to `Discord <https://discordapp.com/>`_ as a bot."""
    interface_name = "discord"

    _identity_table = rb.tables.Discord
    _identity_column = "discord_id"

    def __init__(self,
                 loop: aio.AbstractEventLoop,
                 alchemy_cfg: Dict[str, Any],
                 herald_cfg: Dict[str, Any],
                 sentry_cfg: Dict[str, Any],
                 packs_cfg: Dict[str, Any],
                 serf_cfg: Dict[str, Any],
                 **_):
        if discord is None:
            raise ImportError("'discord' extra is not installed")

        super().__init__(loop=loop,
                         alchemy_cfg=alchemy_cfg,
                         herald_cfg=herald_cfg,
                         sentry_cfg=sentry_cfg,
                         packs_cfg=packs_cfg,
                         serf_cfg=serf_cfg)

        self.token = serf_cfg["token"]
        """The Discord bot token."""

        self.Client = self.client_factory()
        """The custom :class:`discord.Client` class that will be instantiated later."""

        self.client = self.Client()
        """The custom :class:`discord.Client` instance."""

        self.voice_players: List[VoicePlayer] = []
        """A :class:`list` of the :class:`VoicePlayer` in use by this :class:`DiscordSerf`."""

        self.Data: Type[rc.CommandData] = self.data_factory()

    def interface_factory(self) -> Type[rc.CommandInterface]:
        # noinspection PyPep8Naming
        GenericInterface = super().interface_factory()

        # noinspection PyMethodParameters,PyAbstractClass
        class DiscordInterface(GenericInterface):
            name = self.interface_name
            prefix = "!"

        return DiscordInterface

    def data_factory(self) -> Type[rc.CommandData]:
        # noinspection PyMethodParameters,PyAbstractClass
        class DiscordData(rc.CommandData):
            def __init__(data,
                         interface: rc.CommandInterface,
                         loop: aio.AbstractEventLoop,
                         message: "discord.Message"):
                super().__init__(interface=interface, loop=loop)
                data.message = message

            async def reply(data, text: str):
                await data.message.channel.send(escape(text))

            async def get_author(data, error_if_none=False):
                user: "discord.Member" = data.message.author
                query = data.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise rc.CommandError("You must be registered to use this command.")
                return result

            async def delete_invoking(data, error_if_unavailable=False):
                await data.message.delete()

        return DiscordData

    async def handle_message(self, message: "discord.Message"):
        """Handle a Discord message by calling a command if appropriate."""
        text = message.content
        # Skip non-text messages
        if not text:
            return
        # Skip non-command updates
        if not text.startswith("!"):
            return
        # Skip bot messages
        author: Union["discord.User"] = message.author
        if author.bot:
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        # Don't use a case-sensitive command name
        command_name = command_text.lower()
        # Find the command
        try:
            command = self.commands[command_name]
        except KeyError:
            # Skip the message
            return
        # Call the command
        log.debug(f"Calling command '{command.name}'")
        with message.channel.typing():
            # Open an alchemy session, if available
            if self.alchemy is not None:
                session = await asyncify(self.alchemy.Session)
            else:
                session = None
            # Prepare data
            data = self.Data(interface=command.interface, loop=self.loop, message=message)
            # Call the command
            await self.call(command, data, parameters)
            # Close the alchemy session
            if session is not None:
                await asyncify(session.close)

    def client_factory(self) -> Type["discord.Client"]:
        """Create a custom class inheriting from :py:class:`discord.Client`."""
        # noinspection PyMethodParameters
        class DiscordClient(discord.Client):
            async def on_message(cli, message: "discord.Message"):
                """Handle messages received by passing them to the handle_message method of the bot."""
                # TODO: keep reference to these tasks somewhere
                self.loop.create_task(self.handle_message(message))

            async def on_ready(cli) -> None:
                """Change the bot presence to ``online`` when the bot is ready."""
                await cli.change_presence(status=discord.Status.online)

        return DiscordClient

    async def run(self):
        await super().run()
        await self.client.login(self.token)
        await self.client.connect()

    def find_channel(self,
                     channel_type: Optional[Type["discord.abc.GuildChannel"]] = None,
                     name: Optional[str] = None,
                     guild: Optional["discord.Guild"] = None,
                     accessible_to: List["discord.User"] = None,
                     required_permissions: List[str] = None) -> Optional["discord.abc.GuildChannel"]:
        """Find the best channel matching all requests.

        In case multiple channels match all requests, return the one with the most members connected.

        Args:
            channel_type: Filter channels by type (select only :class:`discord.VoiceChannel`,
                          :class:`discord.TextChannel`, ...).
            name: Filter channels by name starting with ``name`` (using :meth:`str.startswith`).
                  Note that some channel types don't have names; this check will be skipped for them.
            guild: Filter channels by guild, keep only channels inside this one.
            accessible_to: Filter channels by permissions, keeping only channels where *all* these users have
                           the required permissions.
            required_permissions: Filter channels by permissions, keeping only channels where the users have *all* these
                                  :class:`discord.Permissions`.

        Returns:
            Either a :class:`~discord.abc.GuildChannel`, or :const:`None` if no channels were found."""
        warnings.warn("This function will be removed soon.", category=DeprecationWarning)
        if accessible_to is None:
            accessible_to = []
        if required_permissions is None:
            required_permissions = []
        channels: List[discord.abc.GuildChannel] = []
        for ch in self.client.get_all_channels():
            if channel_type is not None and not isinstance(ch, channel_type):
                continue

            if name is not None:
                try:
                    ch_name: str = ch.name
                    if not ch_name.startswith(name):
                        continue
                except AttributeError:
                    pass

            ch_guild: "discord.Guild" = ch.guild
            if guild is not None and guild != ch_guild:
                continue

            for user in accessible_to:
                member: "discord.Member" = ch.guild.get_member(user.id)
                if member is None:
                    continue
                permissions: "discord.Permissions" = ch.permissions_for(member)
                missing_perms = False
                for permission in required_permissions:
                    if not permissions.__getattribute__(permission):
                        missing_perms = True
                        break
                if missing_perms:
                    continue

            channels.append(ch)

        if len(channels) == 0:
            return None
        else:
            # Give priority to channels with the most people
            def people_count(c: "discord.VoiceChannel"):
                return len(c.members)

            channels.sort(key=people_count, reverse=True)

            return channels[0]

    def find_voice_players(self, guild: "discord.Guild") -> List[VoicePlayer]:
        candidate_players: List[VoicePlayer] = []
        for player in self.voice_players:
            player: VoicePlayer
            if not player.voice_client.is_connected():
                continue
            if guild is not None and guild != player.voice_client.guild:
                continue
            candidate_players.append(player)
        if guild:
            assert len(candidate_players) <= 1
        return candidate_players



import toml
import importlib
import click
from .version import semantic

p = click.echo


@click.command()
@click.option("-c", "--config-filename", default="./config.toml", type=click.Path(exists=True),
              help="The filename of the Royalnet configuration file.")
@click.option("-f", "--file-format", type=str, help="The name of the format that should be generated.")
def run(config_filename, file_format):
    with open(config_filename, "r") as t:
        config: dict = toml.load(t)

    # Import packs
    packs_cfg = config["Packs"]
    pack_names = packs_cfg["active"]
    packs = {}
    for pack_name in pack_names:
        try:
            packs[pack_name] = {
                "commands": importlib.import_module(f"{pack_name}.commands"),
                "events": importlib.import_module(f"{pack_name}.events"),
                "stars": importlib.import_module(f"{pack_name}.stars"),
                "tables": importlib.import_module(f"{pack_name}.tables"),
            }
        except ImportError as e:
            p(f"Skipping `{pack_name}`: {e}", err=True)
            continue

    if file_format == "botfather":
        for pack_name in packs:
            pack = packs[pack_name]
            lines = []

            try:
                commands = pack["commands"].available_commands
            except AttributeError:
                p(f"Pack `{pack}` does not have the `available_commands` attribute.", err=True)
                continue
            for command in commands:
                lines.append(f"{command.name} - {command.description}")

            lines.sort()
            for line in lines:
                p(line)

    else:
        raise click.ClickException("Unknown format")


if __name__ == "__main__":
    run()



from typing import *
import asyncio as aio
import re
import datetime
import uuid
import logging
import websockets
import royalnet.utils as ru
from .package import Package
from .config import Config


log = logging.getLogger(__name__)


class ConnectedClient:
    """The :py:class:`Server`-side representation of a connected :py:class:`Link`."""
    def __init__(self, socket: "websockets.WebSocketServerProtocol"):
        self.socket: "websockets.WebSocketServerProtocol" = socket
        self.nid: Optional[str] = None
        self.link_type: Optional[str] = None
        self.connection_datetime: datetime.datetime = datetime.datetime.now()

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.nid}>"

    @property
    def is_identified(self) -> bool:
        """Has the client sent a valid identification package?"""
        return bool(self.nid)

    async def send_service(self, msg_type: str, message: str):
        await self.send(Package({"type": msg_type, "service": message},
                                source="<server>",
                                destination=self.nid))

    async def send(self, package: Package):
        """Send a :py:class:`Package` to the :py:class:`Link`."""
        await self.socket.send(package.to_json_bytes())


class Server:
    def __init__(self, config: Config, *, loop: aio.AbstractEventLoop = None):
        self.config: Config = config
        self.identified_clients: List[ConnectedClient] = []
        self.loop = loop

    def __repr__(self):
        return f"<{self.__class__.__qualname__}>"

    def find_client(self, *, nid: str = None, link_type: str = None) -> List[ConnectedClient]:
        assert not (nid and link_type)
        if nid:
            matching = [client for client in self.identified_clients if client.nid == nid]
            assert len(matching) <= 1
            return matching
        if link_type:
            matching = [client for client in self.identified_clients if client.link_type == link_type]
            return matching or []

    async def listener(self, websocket: "websockets.server.WebSocketServerProtocol", path):
        connected_client = ConnectedClient(websocket)
        # Wait for identification
        identify_msg = await websocket.recv()
        log.debug(f"{websocket.remote_address} identified itself with: {identify_msg}.")
        if not isinstance(identify_msg, str):
            log.warning(f"Failed Herald identification: {websocket.remote_address[0]}:{websocket.remote_address[1]}")
            await connected_client.send_service("error", "Invalid identification message (not a str)")
            return
        identification = re.match(r"Identify ([^:\s]+):([^:\s]+):([^:\s]+)", identify_msg)
        if identification is None:
            log.warning(f"Failed Herald identification: {websocket.remote_address[0]}:{websocket.remote_address[1]}")
            await connected_client.send_service("error", "Invalid identification message (regex failed)")
            return
        secret = identification.group(3)
        if secret != self.config.secret:
            log.warning(f"Invalid Herald secret: {websocket.remote_address[0]}:{websocket.remote_address[1]}")
            await connected_client.send_service("error", "Invalid secret")
            return
        # Identification successful
        connected_client.nid = identification.group(1)
        connected_client.link_type = identification.group(2)
        log.info(f"Joined the Herald: {websocket.remote_address[0]}:{websocket.remote_address[1]}"
                 f" ({connected_client.link_type})")
        self.identified_clients.append(connected_client)
        await connected_client.send_service("success", "Identification successful!")
        log.debug(f"{connected_client.nid}'s identification confirmed.")
        # Main loop
        while True:
            # Receive packages
            raw_bytes = await websocket.recv()
            package: Package = Package.from_json_bytes(raw_bytes)
            log.debug(f"Received package: {package}")
            # Check if the package destination is the server itself.
            if package.destination == "<server>":
                # Do... nothing for now?
                pass
            # Otherwise, route the package to its destination
            # noinspection PyAsyncCall
            self.loop.create_task(self.route_package(package))

    def find_destination(self, package: Package) -> List[ConnectedClient]:
        """Find a list of destinations for the package.

        Parameters:
            package: The package to find the destination of.

        Returns:
            A :class:`list` of :class:`ConnectedClient` to send the package to."""
        # Parse destination
        # Is it nothing?
        if package.destination == "<none>":
            return []
        # Is it all possible destinations?
        if package.destination == "*":
            return self.identified_clients
        # Is it a valid nid?
        try:
            destination = str(uuid.UUID(package.destination))
        except ValueError:
            pass
        else:
            return self.find_client(nid=destination)
        # Is it a link_type?
        return self.find_client(link_type=package.destination)

    async def route_package(self, package: Package) -> None:
        """Executed every time a :class:`Package` is received and must be routed somewhere."""
        destinations = self.find_destination(package)
        log.debug(f"Routing package: {package} -> {destinations}")
        for destination in destinations:
            # This may have some consequences
            specific_package = Package(package.data,
                                       source=package.source,
                                       destination=destination.nid,
                                       source_conv_id=package.source_conv_id,
                                       destination_conv_id=package.destination_conv_id)
            await destination.send(specific_package)

    def serve(self):
        if self.config.secure:
            raise Exception("Secure servers aren't supported yet")
        log.debug(f"Serving on {self.config.url}")
        try:
            self.loop.run_until_complete(self.run())
        except OSError as e:
            log.fatal(f"OSError: {e}")
        self.loop.run_forever()

    async def run(self):
        await websockets.serve(self.listener,
                               host=self.config.address,
                               port=self.config.port,
                               loop=self.loop)

    def run_blocking(self, logging_cfg: Dict[str, Any]):
        ru.init_logging(logging_cfg)
        if self.loop is None:
            self.loop = aio.get_event_loop()
        self.serve()



from sqlalchemy import *
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .users import User
import re


class Matrix:
    __tablename__ = "matrix"

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"))

    @declared_attr
    def user(self):
        return relationship("User", backref="matrix")

    @declared_attr
    def matrix_id(self):
        return Column(String, nullable=False, primary_key=True)

    @property
    def username(self):
        match = re.match("^@(.+):.+$", self.matrix_id)
        result = match.group(1)
        assert result is not None
        return result

    @property
    def homeserver(self):
        match = re.match("^@.+:(.+)$", self.matrix_id)
        result = match.group(1)
        assert result is not None
        return result

    def __repr__(self):
        return f"<Matrix {str(self)}>"

    def __str__(self):
        return f"[c]matrix:{self.matrix_id}[/c]"



from typing import Optional


class Config:
    def __init__(self,
                 name: str,
                 address: str,
                 port: int,
                 secret: str,
                 secure: bool = False,
                 path: str = "/"
                 ):
        if ":" in name:
            raise ValueError("Herald names cannot contain colons (:)")
        self.name = name

        self.address = address

        if port < 0 or port > 65535:
            raise ValueError("No such port")
        self.port = port

        self.secure = secure

        if ":" in secret:
            raise ValueError("Herald secrets cannot contain colons (:)")
        self.secret = secret

        if not path.startswith("/"):
            raise ValueError("Herald paths must start with a slash (/)")
        self.path = path

    @property
    def url(self):
        return f"ws{'s' if self.secure else ''}://{self.address}:{self.port}{self.path}"

    def copy(self,
             name: Optional[str] = None,
             address: Optional[str] = None,
             port: Optional[int] = None,
             secret: Optional[str] = None,
             secure: Optional[bool] = None,
             path: Optional[str] = None):
        """Create an exact copy of this configuration, but with different parameters."""
        return self.__class__(name=name if name else self.name,
                              address=address if address else self.address,
                              port=port if port else self.port,
                              secret=secret if secret else self.secret,
                              secure=secure if secure else self.secure,
                              path=path if path else self.path)

    def __repr__(self):
        return f"<HeraldConfig for {self.url}>"

    @classmethod
    def from_config(cls, *,
                    name: str,
                    address: str,
                    port: int,
                    secret: str,
                    secure: bool = False,
                    path: str = "/",
                    enabled: ... = ...
                    ):
        return cls(
            name=name,
            address=address,
            port=port,
            secret=secret,
            secure=secure,
            path=path
        )



from royalnet.utils import *
from royalnet.backpack.tables import *
from royalnet.constellation.api import *


class ApiUserGetStar(ApiStar):
    path = "/api/user/get/v1"

    summary = "Get a Royalnet user by its id."

    parameters = {
        "id": "The id of the user to get."
    }

    tags = ["user"]

    async def api(self, data: ApiData) -> dict:
        user_id_str = data["id"]
        try:
            user_id = int(user_id_str)
        except (ValueError, TypeError):
            raise InvalidParameterError("'id' is not a valid int.")
        user: User = await asyncify(data.session.query(self.alchemy.get(User)).get, user_id)
        if user is None:
            raise NotFoundError("No such user.")
        return user.json()



from starlette.responses import *
from royalnet.utils import *
from royalnet.backpack.tables import *
from royalnet.constellation.api import *


class ApiUserListStar(ApiStar):
    path = "/api/user/list/v1"

    summary = "Get a list of all registered users."

    tags = ["user"]

    async def api(self, data: ApiData) -> dict:
        users: typing.List[User] = await asyncify(data.session.query(self.alchemy.get(User)).all)
        return [user.json() for user in users]



def escape(string: str) -> str:
    """Escape a string to be sent through Telegram (as HTML), and format it using RoyalCode.

    Warning:
        Currently escapes everything, even items in code blocks."""
    return string.replace("<", "&lt;") \
        .replace(">", "&gt;") \
        .replace("[b]", "<b>") \
        .replace("[/b]", "</b>") \
        .replace("[i]", "<i>") \
        .replace("[/i]", "</i>") \
        .replace("[u]", "<b>") \
        .replace("[/u]", "</b>") \
        .replace("[c]", "<code>") \
        .replace("[/c]", "</code>") \
        .replace("[p]", "<pre>") \
        .replace("[/p]", "</pre>")



from typing import *
try:
    from starlette.responses import JSONResponse
except ImportError:
    JSONResponse = None


def api_response(data: dict, code: int, headers: dict = None) -> JSONResponse:
    if headers is None:
        headers = {}
    full_headers = {
        **headers,
        "Access-Control-Allow-Origin": "*",
    }
    return JSONResponse(data, status_code=code, headers=full_headers)


def api_success(data: dict) -> JSONResponse:
    result = {
        "success": True,
        "data": data
    }
    return api_response(result, code=200)


def api_error(error: Exception, code: int = 500) -> JSONResponse:
    result = {
        "success": False,
        "error_type": error.__class__.__qualname__,
        "error_args": list(error.args),
        "error_code": code,
    }
    return api_response(result, code=code)



from typing import *
from sqlalchemy import Column, \
                       Integer, \
                       String, \
                       ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
import royalnet.utils as ru


class Alias:
    __tablename__ = "aliases"

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"), primary_key=True)

    @declared_attr
    def alias(self):
        return Column(String, primary_key=True)

    @declared_attr
    def user(self):
        return relationship("User", backref="aliases")

    @classmethod
    async def find_user(cls, alchemy, session, alias: Union[str, int]):
        result = await ru.asyncify(session.query(alchemy.get(cls)).filter_by(alias=alias.lower()).one_or_none)
        if result is not None:
            result = result.user
        return result

    def __init__(self, user: str, alias: str):
        self.user = user
        self.alias = alias.lower()

    def __repr__(self):
        return f"<Alias {str(self)}>"

    def __str__(self):
        return f"{self.alias}->{self.user_id}"



import royalnet
from royalnet.commands import *


class VersionCommand(Command):
    name: str = "version"

    description: str = "Display the current Royalnet version."

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        # noinspection PyUnreachableCode
        if __debug__:
            message = f"ℹ️ Royalnet {royalnet.__version__} (debug)\n"
        else:
            message = f"ℹ️ Royalnet {royalnet.__version__}\n"
        if "69" in message:
            message += "(Nice.)"
        await data.reply(message)
